// Copyright (c) 1997 Andrew Maltsev <am@amsoft.ru>
// All rights reserved
// 
// EasySQL driver's manager.
//
// See License for your rights and obligations. Or mail to author.
//
#include "es_int.h"
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <errno.h>
#include <ctype.h>

///////////////////////////////////////////////////////////////////////////////
// Initialization and loading of driver
//
easysql::easysql(char const *iname)
        :easysql_error_reporting(iname && *iname ? iname : "easysql")
        ,driver(NULL)
        ,qreg(NULL)
        ,drvname(NULL)
        ,to_be_escaped(NULL)
        ,args(NULL)
        ,retry_delay(DefaultRetryDelay)
{ getConfig();

// Preparing place for arguments
//
  args=new easysql_arguments();
  if(!args)
   { set_error("Cannot create easysql_arguments instance");
     return;
   }

// Name ok?
//
  if(!iname || !*iname)
   { set_error("No database name given");
     return;
   }

// Trying to find database description with given name. System-wide
// databases first.
//
  struct stat st;
  char const *dbname=qprintf("%s/%s",dbDir,iname);
  if(stat(dbname,&st))
   { log(D_DRV|D_L4,"No system database `%s'",iname);
     dbname=NULL;
     if(allowUserConfigs)
      { struct passwd *pe=getpwuid(getuid());		// Using real uid
        dbname=qprintf("%s/.easysql/%s",pe->pw_dir,iname);
        if(stat(dbname,&st))
         { log(D_DRV|D_L4,"No user database `%s'",iname);
           dbname=NULL;
         }
      }
     if(!dbname)
      drvname=strDup(iname);
   }

// Reading description, if found
//
  if(!drvname)
   { if(! st.st_mode&S_IFREG)
      { set_error("%s is not a file",dbname);
        return;
      }
     FILE *f=fopen(dbname,"r");
     if(!f)
      { set_error("Cannot read `%s': %s",dbname,strerror(errno));
        return;
      }
     char buf[200];
     while(fgets(buf,200,f))
      { char *cmd=buf;
        while(isspace(*cmd)) cmd++;
        if(*cmd=='#' || *cmd=='\n' || !*cmd) continue;
        char *val=cmd;
        while(*val && !isspace(*val)) val++;
        if(*val)
         { *val++=0;
           while(*val && isspace(*val)) val++;
         }
        if(!*val)				// We have no one-word commands
         { log(D_DRV|D_L4,"%s: Parameter without value (%s), ignored",iname,cmd);
           continue;
         }
        char *s=val+strlen(val)-1;
        while(s>val && (isspace(*s) || *s=='\n')) *s--=0;
        log(D_DRV|D_L4,"%s: cmd=`%s' val=`%s'",iname,cmd,val);

      // It's handy to have s pointing to the space after first argument
      //
        for(s=val; *s && !isspace(*s); s++);

      // Analysing commands
      //
        if(!strcmp(cmd,"driver"))
         { *s=0;
           drvname=strDup(val);
           log(D_MISC|D_L3,"driver set to `%s'",drvname);
         }
        else if(!strcmp(cmd,"retry-delay"))
         { *s=0;
           retry_delay=strtoul(val,NULL,10);
           log(D_MISC|D_L3,"retry-delay set to %d",retry_delay);
         }
        else
         { if(option(cmd,val))
            log(D_MISC|D_L3,"option `%s' set to `%s'",cmd,val);
           else
            log(D_ERROR|D_MISC,"can't set option `%s' to `%s'",cmd,val);
         }
      }
     fclose(f);
     if(!drvname)
      { set_error("No driver name defined in `%s'",dbname);
        return;
      }
   }

// Creating driver
//
  driver=new easysql_driver_shell(drvname);
  if(!driver)
   { set_error("Cannot create driver `%s'. No memory?",drvname);
     return;
   }
  if(!*driver)
   { set_error("`%s' driver error -- %s",drvname,driver->error_text());
     delete driver;
     driver=NULL;
     return;
   }

// Preparing queries register/reminder
//
  qreg=new easysql_queries_register(driver);
  if(!qreg)
   { set_error("Cannot create queries register. No memory?");
     return;
   }

  no_error();
}

///////////////////////////////////////////////////////////////////////////////
// Finishing driver
//
easysql::~easysql()
{ if(qreg)
   delete qreg;
  if(driver)
   delete driver;
  if(args)
   delete args;
}

///////////////////////////////////////////////////////////////////////////////
// Driver has been successfully loaded/installed?
//
bool
easysql::driver_ok()
{ if(driver)
   { if(! (*driver)) return false;	// Error message will be set in driver
     no_error();
     return true;
   }
  set_error("No driver installed");
  return false;
}

///////////////////////////////////////////////////////////////////////////////
// Checking, that all internal data structures are Ok and interface is
// in operable state
//
easysql::operator bool()
{ if(!driver_ok())
   return false;		// Message already set
  if(!args)
   { set_error("Problems with `args'. No memory?");
     return false;
   }
  if(!qreg)
   { set_error("Problems with `qreg'. No memory?");
     return false;
   }
  return true;
}

///////////////////////////////////////////////////////////////////////////////
// Description of last error. If there is no one then driver's error
// description used. NULL returned in case of no error.
//
char const *
easysql::error_text()
{ char const *e=easysql_error_reporting::error_text();
  if(!e && driver)
   e=driver->error_text();
  return e;
}

///////////////////////////////////////////////////////////////////////////////
// Splitting additional parameters from one input line of form "a=b c=d e=f .."
//
bool
easysql::option(char const *opt)
{ no_error();
  if(driver) drop();			// When called from constructor..
  if(!args || !args->merge(opt))
   return false;
  return true;
}

///////////////////////////////////////////////////////////////////////////////
// Adding parameter by name/value pair. Name and/or value may be NULL, no
// action taken in that case.
//
bool
easysql::option(char const *name, char const *value)
{ no_error();
  if(driver) drop();
  if(!args || !args->merge(name,value))
   return false;
  return true;
}

///////////////////////////////////////////////////////////////////////////////
// Connecting to server
//
bool
easysql::connect()
{ if(connection_ok())				// We are already connected?
   return true;
  if(!(*this))					// driver & args & qreg ok?
   return false;

  if(retry_delay)
   { while(! driver->connect(*args))
      { log(D_ERROR|D_DRV,"Cannot connect: %s",driver->error_text());
        sleep(retry_delay);
      }
     return true;
   }

  return driver->connect(*args);
}

///////////////////////////////////////////////////////////////////////////////
// Connection check
//
bool
easysql::connection_ok()
{ return (*this) && driver->check_connection();
}

///////////////////////////////////////////////////////////////////////////////
// Sending query
//
esqlResult
easysql::query(char const *q, ...)
{ va_list ap;
  va_start(ap,q);
  esqlResult r=_query(q,ap);
  va_end(ap);
  return r;
}

///////////////////////////////////////////////////////////////////////////////
// Sending query from va_list
//
esqlResult
easysql::_query(char const *q, va_list ap)
{ if(!q || !*q)
   { set_error("Empty query string!");
     return NULL;
   }
  if(!connect())
   return NULL;
  no_error();
  char *str=NULL;
  if(strchr(q,'%'))		// Are there any format specifiers?
   str=vaprintf(q,ap);
  esqlResult r;
  while(! driver->query(r,str ? str : q))
   { if(retry_delay)
      { log(D_ERROR|D_DRV,"Cannot send query: %s",driver->error_text());
        sleep(retry_delay);
      }
     else
      { set_error("Cannot send query(\"%s\"): %s",q,driver->error_text());
        r=NULL;
        break;
      }
   }
  if(str)
   Free(str);
  if(r)
   qreg->add(r);
  return r;
}

///////////////////////////////////////////////////////////////////////////////
// Freeing query results
//
bool
easysql::rfree(esqlResult r)
{ if(!driver_ok())
   return false;
  if(!r)
   { set_error("Attempt to rfree() NULL results");
     return false;
   }
  if(!qreg->drop(r))
   { set_error("Bad data given to rfree() - %p",r);
     return false;
   }
  return driver->rfree(r);
}

///////////////////////////////////////////////////////////////////////////////
// Fetching next data row from query results
//
esqlRow
easysql::row(esqlResult r)
{ no_error();
  if(!driver_ok())
   return NULL;
  if(!r)
   { set_error("Attempt to get data from NULL results");
     return NULL;
   }
  esqlRow rr;
  return driver->row(rr,r) ? rr : (esqlRow)NULL;
}

///////////////////////////////////////////////////////////////////////////////
// Dropping connection to engine
//
bool
easysql::drop()
{ if(!connection_ok())
   return false;
  no_error();
  return driver->drop();
}

///////////////////////////////////////////////////////////////////////////////
// Free style driver status
//
char const *
easysql::driver_status()
{ return driver_ok() ? driver->status() : "No driver installed";
}

///////////////////////////////////////////////////////////////////////////////
// Version of driver, API and engine in form of one string:
//  driver_name driver_version [API_version [engine_version]]
//
// API_version or engine_version (or both) may be empty on `unknown'
//
char const *
easysql::driver_version()
{ return driver_ok() ? driver->version() : "unknown";
}

///////////////////////////////////////////////////////////////////////////////
// Setting retry delay. If zero - no retries performed
//
unsigned
easysql::set_retry_delay(unsigned d)
{ unsigned r=retry_delay;
  retry_delay=d;
  if(d)
   log(D_DRV|D_L4,"Retry time set to %u secs",d);
  else
   log(D_DRV|D_L4,"Retries switched off, failures will be reported");
  return r;
}

///////////////////////////////////////////////////////////////////////////////
// Resetting driver's state to default and closing the connection
//
void
easysql::reset()
{ drop();
  args->reset();
}

//////////////////////////////////////////////////////////////////////////////
// Converting text string to SQL-ready format:
//  abc"'\def   ->   abc\"\'\\def
// Symbols to be escaped are requested from driver on first call
//
// If pbuf!=NULL, then it is used as storage space for converted string. Else internal buffer is
// used (and IS OVERRIDEN on every call!)
//
char const *
easysql::string(char const *str, char *pbuf)
{ static char *buf=NULL;
  static unsigned size=0;

// If driver is not ready no conversion is done. It shouldn't happen!
//
  if(!driver_ok())
   return str;

// What symbols have to be escaped?
//
  if(!to_be_escaped)
   { to_be_escaped=driver->escaped_chars();
     if(!to_be_escaped || !*to_be_escaped || !to_be_escaped[1])
      { log(D_ERROR|D_MISC,"Wrong `to_be_escaped' string returned by %s driver",drvname);
        to_be_escaped="\\'\"";
      }
     if(*to_be_escaped!='\\')
      log(D_MISC|D_L1,"First symbol of `to_be_escaped' string (%s) is wrong (%s driver)"
                     ,to_be_escaped,drvname);
   }

// Driver wants to do conversion by itself?
//
//  if(!to_be_escaped || !*to_be_escaped)
//   return driver->string(str,pbuf);

// Selecting buffer and reallocating it if required
//
  char *b;
  if(pbuf)
   b=pbuf;		// Using outside buffer
  else
   { if(!str)		// using internal dynamic buffer
      str="<NULL>";
     if(strlen(str)*2>=size)
      buf=(typeof(buf))reAlloc(buf,sizeof(char)*(size=strlen(str)*2+1));
     b=buf;
   }

// Escaping
//
  char *s=b;
  for(; *str; str++, s++)
   { if(index(to_be_escaped,*str))
      *s++=*to_be_escaped;
     *s=*str;
   }
  *s=0;
  return b;
}

///////////////////////////////////////////////////////////////////////////////
R("$Id: easysql.C,v 1.11 1997/11/05 19:51:56 am Exp $");
