// Copyright (c) 1997 Andrew Maltsev <am@amsoft.ru>
// All rights reserved
// 
// Driver's shell.
//
// Loads shered object from disk on in-core register, checks it. Then
// gets entry point for driver's functions and builds a shell class more
// suitable for further work.
//
#include <stdio.h>
#include <stdarg.h>
#include "es_int.h"

///////////////////////////////////////////////////////////////////////////////
// Driver loader
//
easysql_driver_loader::easysql_driver_loader()
                      :driver(NULL)
{
}
easysql_driver_loader::~easysql_driver_loader()
{
}
void *
easysql_driver_loader::operator() (char const *par) const
{ return driver ? driver(par) : NULL;
}
easysql_driver_loader::operator bool() const
{ return driver;
}

///////////////////////////////////////////////////////////////////////////////
// Loading driver from on disk shared object
//
class easysql_disk_loader : public easysql_driver_loader
{ easysql_dynamic_loader loader;
 public:
  easysql_disk_loader(char const *);
  ~easysql_disk_loader();
};

///////////////////////////////////////////////////////////////////////////////
// It for now looks only in system driver's folder. Later may be extended
// to allow per user folders
//
easysql_disk_loader::easysql_disk_loader(char const *name)
                    :easysql_driver_loader()
{ loader.load(qprintf("%s/%s.so",driversDir,name));
  if(!loader)
   return;
  driver=(typeof(driver))loader.symbol(qprintf("%s_easysql_driver",name));
}

///////////////////////////////////////////////////////////////////////////////
// Shared object will be freed by destruction of loader automatically
//
easysql_disk_loader::~easysql_disk_loader()
{
}

///////////////////////////////////////////////////////////////////////////////
// Loader from system in-core register
//
class easysql_registry_loader : public easysql_driver_loader
{public:
  easysql_registry_loader(char const *);
  ~easysql_registry_loader();
};

easysql_registry_loader::easysql_registry_loader(char const *name)
                        :easysql_driver_loader()
{ driver=easysql_register::find(name);
}

easysql_registry_loader::~easysql_registry_loader()
{
}

///////////////////////////////////////////////////////////////////////////////
// Tries different loaders and returns a reference to a class instance
// in which the search succeeded.
//
static
easysql_driver_loader &
load_module(char const *name)
{ easysql_driver_loader *drv;
  drv=new easysql_registry_loader(name);
  if(!drv || !(*drv))
   { if(drv) delete drv;
     drv=new easysql_disk_loader(name);
     if(!drv || !(*drv))
      { if(drv) delete drv;
        drv=NULL;
      }
   }
  return *drv;
}

///////////////////////////////////////////////////////////////////////////////
// Driver shell initializing
//
easysql_driver_shell::easysql_driver_shell(char const *name)
                     :easysql_error_reporting("driver_shell",D_DRV)
                     ,drv(load_module(name))
                     ,data(NULL)
                     ,i_version(NULL)
                     ,i_create(NULL),i_finish(NULL)
                     ,i_connect(NULL)
                     ,i_query(NULL)
                     ,i_rfree(NULL)
                     ,i_row(NULL)
                     ,i_drop(NULL)
                     ,i_error(NULL),i_no_error(NULL)
                     ,i_status(NULL),i_check_connection(NULL)
                     ,i_to_be_escaped(NULL)
{ if(&drv==NULL || !drv)
   return;

// Testing for `void *' and `char' size
//
  unsigned long *pul=(typeof(pul))drv(NULL);
  if(!pul)
   { set_error("%s: no pointer size returned from driver",name);
     return;
   }
  if((*pul)%32!=sizeof(void *))
   { set_error("%s: ptr size mismatch -- %lu!=%lu\n",name,(*pul)%32,(unsigned long)(sizeof(void*)));
     return;
   }
  if((*pul)/32!=sizeof(char))
   { set_error("%s: char size mismatch -- %lu!=%lu\n",name,(*pul)/32,(unsigned long)(sizeof(void*)));
     return;
   }

// Testing conversion rules
//
  void *pv=drv("_54321");
  if(pv!=(void *)54321)
   { set_error("%s: ptr conversion rules mismatch -- %p!=%p\n",name,pv,(void *)54321);
     return;
   }

// Testing protocol number. Later it will be more complex (possible),
// but for now it is just comparision for equality
//
  pul=(typeof(pul))drv("_protocol");
  if(!pul)
   { set_error("%s: no protocol type returned from driver",name);
     return;
   }
  if(*pul != EasySQL_ProtocolVersion)
   { set_error("%s: driver protocol version mismatch -- %lu!=%u"
             ,name,*pul,(unsigned)EasySQL_ProtocolVersion);
     return;
   }

// Filling shell with real functions
//
  i_version		=(typeof(i_version))		drv("version");
  i_create		=(typeof(i_create))		drv("create");
  i_finish		=(typeof(i_finish))		drv("finish");
  i_connect		=(typeof(i_connect))		drv("connect");
  i_drop		=(typeof(i_drop))		drv("drop");
  i_query		=(typeof(i_query))		drv("query");
  i_row			=(typeof(i_row))		drv("row");
  i_rfree		=(typeof(i_rfree))		drv("rfree");
  i_error		=(typeof(i_error))		drv("error");
  i_no_error		=(typeof(i_no_error))		drv("no_error");
  i_status		=(typeof(i_status))		drv("status");
  i_check_connection	=(typeof(i_check_connection))	drv("check_connection");
  i_to_be_escaped	=(typeof(i_to_be_escaped))	drv("to_be_escaped");

// Creating driver's instance data
//
  if(!i_create)
   { set_error("No `%s' driver instance creator found",name);
     return;
   }
  data=i_create();
  if(!data)
   { set_error("Cannot create `%s' driver's data instance",name);
     return;
   }
  no_error();
}

///////////////////////////////////////////////////////////////////////////////
// Finishing shell and driver
//
easysql_driver_shell::~easysql_driver_shell()
{ if(data && i_finish)
   i_finish(data);
  delete &drv;
}

///////////////////////////////////////////////////////////////////////////////
// Checking for consistency
//
easysql_driver_shell::operator bool()
{ no_error();
  bool r=&drv &&
         drv &&
         data &&
         i_create && i_finish &&
         i_version && i_connect && i_query && i_rfree && i_row && i_drop &&
         i_error && i_no_error &&
         i_status && i_check_connection;
  if(!r)
   set_error("Driver not completelly installed. Incompatible?");
  return r;
}

///////////////////////////////////////////////////////////////////////////////
// Returns text of last error occured -- both in driver and in shell
//
char const *
easysql_driver_shell::error_text()
{ char const *e=easysql_error_reporting::error_text();
  if(!e && data && i_error)
   e=i_error(data);
  return e;
}

///////////////////////////////////////////////////////////////////////////////
// Clearing error text -- our self and driver's
//
void
easysql_driver_shell::no_error()
{ easysql_error_reporting::no_error();
  if(data && i_no_error)
   i_no_error(data);
}

#ifdef TEST
///////////////////////////////////////////////////////////////////////////////
// Test-Suite
//
int
main(int ac, char *av[])
{ char const *module=(ac==2 ? av[1] : "null");
  printf("=================== %s =====================\n",module);
  easysql_driver_shell drv(module);
  if(!drv)
   { printf("Can't load driver\n");
     exit(1);
   }
  printf("Version: %s\n",drv.version());

  return 0;
}
#endif

///////////////////////////////////////////////////////////////////////////////
R("$Id: shell.C,v 1.4 1997/10/23 18:40:43 am Exp $");
